@@ -21,15 +21,15 @@ import numpy as np
 
 class NetCDFOutput:
 
-    def __init__( self, filename=None, M=8, N=8 ):
+    def __init__( self, filename=None, M=0, N=0 ):
 
         self.rootgrp = None
         if filename == None:
             return
 
         self.rootgrp = Dataset(filename, "w", format="NETCDF4")
-        self.rootgrp.createDimension("X")
-        self.rootgrp.createDimension("Y")
+        self.rootgrp.createDimension("X", N)
+        self.rootgrp.createDimension("Y", M)
         self.rootgrp.createDimension("count")
         self.metagrp = self.rootgrp.createGroup("meta")
 
@@ -72,12 +72,12 @@ class NetCDFOutput:
         self.ky.long_name = "Vertical wavenumbers"
         self.ky.field = "Ky, scalar, series"
 
-        self.Z = self.rootgrp.createVariable( "Z", "f4", ("count", "X","Y",)) #, chunksizes=(1,M//8,N//8) )
+        self.Z = self.rootgrp.createVariable( "Z", "f4", ("count", "X","Y") )
         self.Z.units = "millimeter"
         self.Z.long_name = "Z data on time over the XY grid"
         self.Z.field = "Z, scalar, series"
 
-        self.maskZ = self.rootgrp.createVariable( "maskZ", "f4", ("X","Y",) )
+        self.maskZ = self.rootgrp.createVariable( "maskZ", "f4", ("X","Y") )
         self.maskZ.units = ""
         self.maskZ.long_name = "Z mask over the XY grid"
         self.maskZ.field = "Z, scalar, series"
@@ -113,7 +113,7 @@ class NetCDFOutput:
         self.maskZ[:] = maskZ
 
 
-    def set_instrinsics( self, K0, K1, kc0, kc1, P0plane ):
+    def set_instrinsics( self, K0, K1, kc0, kc1, P0plane, P1plane, P0cam, P1cam, Cam0toGrid, Cam1toGrid ):
         if self.rootgrp == None:
             return
 
@@ -122,12 +122,37 @@ class NetCDFOutput:
         self.metagrp.createDimension( "DistV", 5 )
         K0v = self.metagrp.createVariable("intr0", "f8", ("V3","V3") )
         K0v[:] = K0
+
         K1v = self.metagrp.createVariable("intr1", "f8", ("V3","V3") )
         K1v[:] = K1
-        _P0plane = self.metagrp.createVariable("P0plane", "f8", ("V4","V4") )
-        _P0plane[:] = P0plane
+
+        if not P0plane is None:
+            _P0plane = self.metagrp.createVariable("P0plane", "f8", ("V4","V4") )
+            _P0plane[:] = P0plane
+
+        if not P1plane is None:
+            _P1plane = self.metagrp.createVariable("P1plane", "f8", ("V4","V4") )
+            _P1plane[:] = P1plane
+
+        if not P0cam is None:
+            _P0cam = self.metagrp.createVariable("P0cam", "f8", ("V3","V4") )
+            _P0cam[:] = P0cam
+
+        if not P1cam is None:
+            _P1cam = self.metagrp.createVariable("P1cam", "f8", ("V3","V4") )
+            _P1cam[:] = P1cam
+
+        if not Cam0toGrid is None:
+            _Cam0toGrid = self.metagrp.createVariable("Cam0toGrid", "f8", ("V4","V4") )
+            _Cam0toGrid[:] = Cam0toGrid
+
+        if not Cam1toGrid is None:
+            _Cam1toGrid = self.metagrp.createVariable("Cam1toGrid", "f8", ("V4","V4") )
+            _Cam1toGrid[:] = Cam1toGrid
+
         kc0v = self.metagrp.createVariable("dist0", "f8", ("DistV") )
         kc0v[:] = kc0
+
         kc1v = self.metagrp.createVariable("dist1", "f8", ("DistV") )
         kc1v[:] = kc1
 
@@ -140,9 +165,11 @@ class NetCDFOutput:
     #         if idx%10 == 0:
     #             self.rootgrp.sync() # Flush data to disk
 
-    def push_Z( self, Zdata, time, workdir, image=None, imagemask=None ):
+    def push_Z( self, Zdata, time, workdir, image=None, imagemask=None, idx=None ):
         if self.rootgrp != None:
-            idx = self.count.shape[0]
+            if idx is None:
+                idx = self.count.shape[0]
+
             self.Z[idx,:,:] = np.expand_dims( Zdata, axis=0 )
             self.count[idx] = idx
             self.time[idx] = time
@@ -154,8 +181,8 @@ class NetCDFOutput:
             if not imagemask is None:
                 self.cam0masks[idx] = imagemask
 
-            if idx%10 == 0:
-                self.rootgrp.sync() # Flush data to disk
+            #if idx%10 == 0:
+            #    self.rootgrp.sync() # Flush data to disk
 
 
     # def add_global_attribute( self, name, value ):
