@@ -1,6 +1,6 @@
 """
 wassgridsurface
-Copyright (C) 2023 Filippo Bergamasco
+Copyright (C) 2025 Filippo Bergamasco
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -17,8 +17,11 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
 
-VERSION = "0.5.7"
+VERSION = "0.11.4"
 
+import matplotlib
+matplotlib.use('AGG')
+import matplotlib.pyplot as plt
 
 import argparse
 import configparser
@@ -27,11 +30,13 @@ import scipy.interpolate
 import os
 import sys
 import glob
+import time
 import numpy as np
-import matplotlib.pyplot as plt
 import cv2 as cv
 from os import path
-from tqdm import tqdm
+from tqdm import tqdm, trange
+from tqdm.contrib.concurrent import thread_map
+
 
 import colorama
 colorama.init()
@@ -48,8 +53,13 @@ from wassgridsurface.DCTInterpolator import DCTInterpolator
 
 
 def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_size_y, Nx, Ny, Iw=None, Ih=None, fps=0, timestring="" ):
-    meshname = path.join(wdir,"mesh_cam.xyzC")
-    print("Loading ", meshname )
+
+    s = cv.FileStorage()
+
+    s.open(path.join(wdir,"intrinsics_00000000.xml"), cv.FileStorage_READ)
+    K0 = s.getNode('intr').mat() 
+    s.open(path.join(wdir,"intrinsics_00000001.xml"), cv.FileStorage_READ)
+    K1 = s.getNode('intr').mat() 
 
     R = np.loadtxt(path.join(wdir,'Cam0_poseR.txt'))
     T = np.loadtxt(path.join(wdir,'Cam0_poseT.txt'))
@@ -68,6 +78,9 @@ def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_siz
     print("Camera image size: "+Fore.RED+("%dx%d"%(Iw,Ih))+Fore.RESET+" (override with -Iw -Ih program arguments)")
 
     Rpl, Tpl = compute_sea_plane_RT( meanplane )
+
+    meshname = path.join(wdir,"mesh_cam.xyzC")
+    print("Loading ", meshname )
     mesh = load_camera_mesh(meshname)
     mesh_aligned = align_on_sea_plane( mesh, meanplane) * baseline
 
@@ -81,10 +94,25 @@ def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_siz
                         [ 0,      0,       0, 1]], dtype=float )
 
     SCALEi = 1.0/baseline
-    P0plane = toNorm @ P0Cam @ RTplane @ np.diag((SCALEi,SCALEi,-SCALEi, 1))
+    D = np.diag((SCALEi,SCALEi,-SCALEi, 1))
 
-    area_size_m_x = np.floor( area_size_x / 2)
-    area_size_m_y = np.floor( area_size_y / 2)
+    P0plane = toNorm @ P0Cam @ RTplane @ D
+    P1plane = toNorm @ P1Cam @ RTplane @ D
+
+    K0_44 = np.eye(4)
+    K0_44[:3,:3] = K0
+    K1_44 = np.eye(4)
+    K1_44[:3,:3] = K1
+
+    # Compute 4x4 matrices transforming points from
+    # camera reference system to grid reference system
+    #
+    Cam0toGrid = np.linalg.inv(D) @ np.linalg.inv(RTplane) @ np.linalg.inv( np.linalg.inv(K0_44) @ P0Cam )
+    Cam1toGrid = np.linalg.inv(D) @ np.linalg.inv(RTplane) @ np.linalg.inv( np.linalg.inv(K1_44) @ P1Cam )
+
+
+    area_size_m_x = area_size_x / 2
+    area_size_m_y = area_size_y / 2
     xmin = area_center[0]-area_size_m_x
     xmax = area_center[0]+area_size_m_x
     ymin = area_center[1]-area_size_m_y
@@ -100,6 +128,33 @@ def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_siz
 
     print("zmin .. zmax = %3.2f ... %3.2f"%(zmin,zmax) )
 
+    # Plane-image homographies
+    gridbounds = np.array( [[xmin,ymin,0,1],[xmax,ymin,0,1],[xmax,ymax,0,1],[xmin,ymax,0,1]], dtype=float ).T
+    gridbounds_cam0 = np.linalg.inv(toNorm) @ P0plane @ gridbounds
+    gridbounds_cam0 /= gridbounds_cam0[2,:]
+
+    gridbounds_cam1 = np.linalg.inv(toNorm) @ P1plane @ gridbounds
+    gridbounds_cam1 /= gridbounds_cam1[2,:]
+
+    lineto = lambda I,ii,jj : cv.line( I, (int(gridbounds_cam0[0,ii]),int(gridbounds_cam0[1,ii])), (int(gridbounds_cam0[0,jj]),int(gridbounds_cam0[1,jj])), (0,0,255), 4 )
+
+    I = cv.imread(  path.join(wdir,"undistorted","00000000.png"), cv.IMREAD_GRAYSCALE )
+    I = cv.cvtColor(I, cv.COLOR_GRAY2BGR)
+
+    gridbounds_texture = np.array( [[0,0],[Nx,0],[Nx,Ny],[0,Ny]], dtype=np.float32 ).T
+    Hcam0_to_grid, _ = cv.findHomography( gridbounds_cam0[:2,:].astype(np.float32).T, gridbounds[:2,:].astype(np.float32).T )
+    Hcam1_to_grid, _ = cv.findHomography( gridbounds_cam1[:2,:].astype(np.float32).T, gridbounds[:2,:].astype(np.float32).T )
+    Hcam0_to_texture, _ = cv.findHomography( gridbounds_cam0[:2,:].astype(np.float32).T, gridbounds_texture.T )
+
+    I0rect = cv.warpPerspective( I, Hcam0_to_texture, (Nx,Ny) )
+    cv.imwrite( path.join(outdir,"cam0_rectified.jpg"), I0rect )
+
+    lineto(I,0,1)
+    lineto(I,1,2)
+    lineto(I,2,3)
+    lineto(I,3,0)
+    cv.imwrite( path.join(outdir,"grid_projected_cam0.jpg"),I )
+
     # Meshgrid
     XX,YY = np.meshgrid( np.linspace(xmin,xmax,Nx), np.linspace(ymin,ymax,Ny) )
     x_spacing = XX[0,1]-XX[0,0]
@@ -143,14 +198,23 @@ def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_siz
         "zmax":zmax,
         "P0cam":P0Cam[0:3,:],
         "P1cam":P1Cam[0:3,:],
+        "Hcam0toGrid":Hcam0_to_grid,
+        "Hcam1toGrid":Hcam1_to_grid,
+        "Cam0toGrid":Cam0toGrid,
+        "Cam1toGrid":Cam1toGrid,
+        "Hcam0toTexture":Hcam0_to_texture,
         "Nx":Nx,
         "Ny":Ny,
         "N":np.amax([Nx,Ny]),
         "R":R,
         "T":T,
+        "RTplane":RTplane,
+        "K0":K0,
+        "K1":K1,
         "Rpl":Rpl,
         "Tpl":Tpl,
         "P0plane":P0plane,
+        "P1plane":P1plane,
         "CAM_BASELINE":baseline,
         "scale":baseline,
         "XX":XX,
@@ -166,8 +230,8 @@ def setup( wdir, meanplane, baseline, outdir, area_center, area_size_x, area_siz
 
 
 
-def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="DCT", user_mask_filename=None, alg_options=None ):
-    step=150
+def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="DCT", user_mask_filename=None, alg_options=None, NUM_PARALLEL_PROCESSES=1, image_id_to_save=0, force_zero_mean=False ):
+
     gridsetup = sio.loadmat( matfile )
     XX = gridsetup["XX"]
     YY = gridsetup["YY"]
@@ -177,33 +241,39 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
     Rp2c = Rpl.T
     Tp2c = -Rp2c@Tpl
 
-    P0cam = gridsetup["P0cam"]
-    P1cam = gridsetup["P1cam"]
 
     outdata = NetCDFOutput( filename=path.join(outdir,"gridded.nc" ), M=XX.shape[0], N=XX.shape[1] )
     baseline = gridsetup["CAM_BASELINE"].item(0)
 
     fps = gridsetup["fps"].item(0)
+    stereo_image_index = image_id_to_save
 
     outdata.scale[:] = baseline
     outdata.add_meta_attribute("info", "Generated with WASS gridder v.%s"%VERSION )
     outdata.add_meta_attribute("generator", "WASS" )
     outdata.add_meta_attribute("baseline", baseline )
     outdata.add_meta_attribute("fps", fps)
+    outdata.add_meta_attribute("stereo_image_index", stereo_image_index)
     outdata.add_meta_attribute("timestring", gridsetup["timestring"] )
     outdata.set_grids( XX*1000.0, YY*1000.0 )
     outdata.set_kxky( gridsetup["KX_ab"], gridsetup["KY_ab"] )
 
-    outdata.set_instrinsics( np.zeros( (3,3), dtype=np.float32),
-                             np.zeros( (3,3), dtype=np.float32),
-                             np.zeros( (5,1), dtype=np.float32),
-                             np.zeros( (5,1), dtype=np.float32),
-                             gridsetup["P0plane"])
 
-    Zmean = 0.0
-    Zmin = np.Inf
-    Zmax = -np.Inf
-    N_frames = 1
+    outdata.set_instrinsics( gridsetup["K0"],
+                             gridsetup["K1"],
+                             np.zeros( (5,1), dtype=np.float32),
+                             np.zeros( (5,1), dtype=np.float32),
+                             gridsetup["P0plane"],
+                             gridsetup["P1plane"],
+                             gridsetup["P0cam"],
+                             gridsetup["P1cam"],
+                             gridsetup["Cam0toGrid"],
+                             gridsetup["Cam1toGrid"])
+
+
+    wass_frames_with_indices = [ x for x in enumerate(wass_frames) ]
+    N_frames = len( wass_frames_with_indices )
+
 
     user_mask = np.ones( XX.shape, dtype=np.float32 )
 
@@ -214,13 +284,30 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
 
 
     print("Interpolation algorithm: "+Fore.RED+algorithm+Fore.RESET )
-    interpolator = IDWInterpolator( KSIZE=5, reps=1 ) if algorithm=="IDW" else DCTInterpolator( img_width=XX.shape[1], img_height=XX.shape[0], alg_options=alg_options )
+    print("Using %d thread(s) for reconstruction..."%NUM_PARALLEL_PROCESSES )
+
+
+    # Create a dedicated interpolator for each parallel process
+    interpolators = [ IDWInterpolator( KSIZE=5, reps=1 ) if algorithm=="IDW" else DCTInterpolator( img_width=XX.shape[1], img_height=XX.shape[0], alg_options=alg_options ) for x in range(NUM_PARALLEL_PROCESSES) ]
+
+    Zmeans = []
+    Zmins = []
+    Zmaxs = []
+
+    Zmean_grids = [ np.zeros_like( XX ) for x in range(NUM_PARALLEL_PROCESSES) ]   # this will store per-pixel average for each interpolator
+
+    def _grid_task( iteration_element ):
+
+        idx, wdir = iteration_element
+        interpolator = interpolators[ idx % NUM_PARALLEL_PROCESSES ]
 
-    for wdir in tqdm(wass_frames):
         tqdm.write(wdir)
         dirname = path.split( wdir )[-1]
         FRAME_IDX = int(dirname[:-3])
 
+        if idx<=NUM_PARALLEL_PROCESSES:
+            time.sleep( idx ) # wait to optimize disk access
+
         meshname = path.join( wdir, "mesh_cam.xyzC")
         mesh = load_camera_mesh(meshname)
         mesh_aligned = align_on_sea_plane_RT( mesh, gridsetup["Rpl"], gridsetup["Tpl"]) * gridsetup["CAM_BASELINE"]
@@ -253,14 +340,14 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
 
             ZZ = np.nanmedian( ZZ, axis=-1 )
 
-            if N_frames == 1:
+            if idx == 0:
                 fig = plt.figure( figsize=(20,20))
                 plt.imshow( ZZ, vmin=gridsetup["zmin"], vmax=gridsetup["zmax"] )
                 figfile = path.join(outdir,"points.png" )
                 fig.savefig(figfile,bbox_inches='tight')
                 plt.close()
 
-            Zi, mask = interpolator(ZZ)
+            Zi, mask = interpolator(ZZ, verbose=(NUM_PARALLEL_PROCESSES==1) )
             mask *= user_mask
             Zi[ mask==0 ] = np.nan
 
@@ -270,7 +357,7 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
                 Zi = cv.medianBlur(Zi, ksize=mf)
                 Zi[ mask==0 ] = np.nan
 
-            if N_frames == 1:
+            if idx == 0:
                 fig = plt.figure( figsize=(20,20))
                 plt.imshow( Zi, vmin=gridsetup["zmin"], vmax=gridsetup["zmax"] )
                 figfile = path.join(outdir,"gridded.png" )
@@ -393,24 +480,31 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
             return
 
 
-        # Zi here contains the interpolated surface
-        Zmean = Zmean + (np.nanmean(Zi)-Zmean)/N_frames
-        Zmin = min( Zmin, np.nanmin(Zi) )
-        Zmax = max( Zmax, np.nanmax(Zi) )
+        # Zi now contains the interpolated surface
 
+        Zmean_grids[ idx % NUM_PARALLEL_PROCESSES ] = Zmean_grids[ idx % NUM_PARALLEL_PROCESSES ] + Zi 
+
+        Zmeans.append( np.nanmean(Zi) )
+        Zmins.append( np.nanmin(Zi) )
+        Zmaxs.append( np.nanmax(Zi) )
+
+        img_to_load = path.join(wdir,"undistorted","%08d.png"%image_id_to_save ) 
+        I0 = cv.imread(img_to_load)
+        if I0.shape[0] == 0:
+            img_to_load = path.join(wdir,"%08d_s.png"%image_id_to_save)
+            I0 = cv.imread(img_to_load)
 
-        I0 = cv.imread( path.join(wdir,"00000000_s.png"))
         outdata.add_meta_attribute("image_width", I0.shape[1] )
         outdata.add_meta_attribute("image_height", I0.shape[0] )
-        ret, imgjpeg = cv.imencode(".jpg", I0 )
+        _, imgjpeg = cv.imencode(".jpg", I0 )
 
-        mask_filename = path.join( wdir, "undistorted", "mask0.png" ) 
+        mask_filename = path.join( wdir, "undistorted", "mask%d.png"%image_id_to_save ) 
         imagemask = None
         if path.exists( mask_filename ): 
             with open( mask_filename, "rb" ) as f:
                 imagemask = np.fromfile(f, np.uint8 );
 
-        outdata.push_Z( Zi*1000, float(N_frames-1)/float(fps) if fps>0 else 0.0, FRAME_IDX, imgjpeg, imagemask )
+        outdata.push_Z( Zi*1000, float(idx)/float(fps) if fps>0 else 0.0, FRAME_IDX, imgjpeg, imagemask, idx=idx )
 
 
         #aux = ((Zi-gridsetup["zmin"])/(gridsetup["zmax"]-gridsetup["zmin"])*255).astype(np.uint8)
@@ -422,7 +516,41 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
         # fig.savefig(figfile,bbox_inches='tight')
         # plt.close()
 
-        N_frames += 1
+    #------
+    r = thread_map(_grid_task, wass_frames_with_indices, max_workers=NUM_PARALLEL_PROCESSES )
+
+    Zmin = np.amin( np.array(Zmins)) 
+    Zmax = np.amax( np.array(Zmaxs)) 
+    Zmean = np.mean( np.array(Zmeans)) 
+
+
+
+    Zmean_perpoint = np.sum( np.array( Zmean_grids), axis=0)  / float(N_frames) * 1000
+    fig = plt.figure( figsize=(20,20))
+    plt.imshow(Zmean_perpoint)
+    plt.colorbar()
+    figfile = path.join(outdir,"average_elevation.png")
+    fig.savefig(figfile,bbox_inches='tight')
+    plt.close()
+
+
+    if force_zero_mean:
+        print("Forcing local surface elevation to zero")
+        for ii in trange( N_frames ):
+            outdata.Z[ ii, :, :] = outdata.Z[ ii, :, :] - Zmean_perpoint
+        Zmean = 0.0
+        Zmax = Zmin
+
+
+#    Znetcdf = np.array( outdata.Z )
+#    print(Znetcdf.shape)
+#    Zmean_perpoint = np.sum( np.array( Znetcdf), axis=0)  / float(N_frames)
+#    fig = plt.figure( figsize=(20,20))
+#    plt.imshow(Zmean_perpoint)
+#    plt.colorbar()
+#    figfile = path.join(outdir,"average_elevation_netcdf.png")
+#    fig.savefig(figfile,bbox_inches='tight')
+#    plt.close()
 
 
     outdata.add_meta_attribute("zmin", Zmin )
@@ -433,7 +561,7 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
     print("    Zmin: ",Zmin)
     print("    Zmax: ",Zmax)
     print("   Zmean: ",Zmean)
-    print("# frames: ",N_frames-1)
+    print("# frames: ",N_frames)
 
     outdata.close()
 
@@ -441,7 +569,7 @@ def grid( wass_frames, matfile, outdir, subsample_percent=100, mf=0, algorithm="
 
 def wassgridsurface_main():
     print(" WASS surface gridder v.", VERSION )
-    print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\nCopyright (C) Filippo Bergamasco 2022 \n")
+    print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\nCopyright (C) Filippo Bergamasco 2022-2025 \n")
 
     howtostring = """
         How to use:
@@ -451,12 +579,12 @@ def wassgridsurface_main():
             mkdir gridding
 
         2) Generate grid config file:
-            wassgridsurface --action generategridconfig . gridding
+            wassgridsurface --action generateconfig ./output ./gridding
 
         3) Edit the grid config file ./gridding/gridconfig.txt with your favourite editor
 
         4) Setup the reconstruction
-            wassgridsurface --action setup ./output ./gridding --gridconfig ./gridding/gridconfig.txt --baseline [CAMERA_BASELINE]
+            wassgridsurface --action setup ./output ./gridding --gridconfig ./gridding/gridconfig.txt --fps [SEQUENCE FPS] --baseline [CAMERA_BASELINE]
 
         5) Open the image ./gridding/area_grid.png to check the extension of the reconstructed area.
            To make changes, go back to step 3.
@@ -470,7 +598,7 @@ def wassgridsurface_main():
     parser = argparse.ArgumentParser( epilog=howtostring, formatter_class=argparse.RawDescriptionHelpFormatter )
     parser.add_argument("workdir", help="WASS output directory containing the reconstructed frames")
     parser.add_argument("outdir", help="Output directory")
-    parser.add_argument("--action", type=str, help="What to do [setup | grid | generategridconfig ]" )
+    parser.add_argument("--action", choices=("setup","grid","generateconfig"), type=str, help="What to do [setup | grid | generateconfig ]" )
     parser.add_argument("--gridconfig", type=str, help="Grid configuration file for setup action")
     parser.add_argument("--gridsetup", type=str, help="Grid setup file for grid action")
     parser.add_argument("-b", "--baseline", default=1.0, type=float, help="Stereo camera distance" )
@@ -481,8 +609,11 @@ def wassgridsurface_main():
     parser.add_argument("--ss", "--subsample_percent", type=float, default=100, help="Point subsampling 0..100%%" )
     parser.add_argument("--mf", "--medianfilter", type=int, default=0, help="Median filter window size (0,3,5,7)" )
     parser.add_argument("-n", "--num_frames", type=int, default=-1, help="Number of frames to process. -1 to process all frames." )
+    parser.add_argument("-p", "--parallel", type=int, default=1, help="Number of parallel tasks to execute" )
     parser.add_argument("--ia", "--interpolation_algorithm", type=str, default="DCT", help='Interpolation algorithm to use. Alternatives are: "DCT", "IDW", "LinearND" ' )
     parser.add_argument("--mask", type=str, default=None, help='User supplied grid mask filename. Must be a grayscale (bw) image with the same size of the grid' )
+    parser.add_argument("-z", "--force-zero-mean", action="store_true", help='Subtracts the per-grid-point average elevation from each surface' )
+    parser.add_argument("--stereo_image_idx", type=int, default=0, help='Which stereo frame to store in the NetCDF, 0 or 1' )
     parser.add_argument("--dct_nfreqs", type=int, default=None, help="DCT interpolator number of frequencies" )
     parser.add_argument("--dct_regalpha", type=float, default=None, help="DCT interpolator regularizer alpha" )
     parser.add_argument("--dct_maxtol", type=float, default=None, help="DCT interpolator max function tolerance change" )
@@ -490,7 +621,7 @@ def wassgridsurface_main():
     parser.add_argument("--dct_maxiters", type=int, default=None, help="DCT interpolator max number of iterations" )
     args = parser.parse_args()
 
-    if args.action == "generategridconfig":
+    if args.action == "generateconfig":
         gridconfigfile = path.join(args.outdir,"gridconfig.txt")
         print("Generating ",gridconfigfile)
         with open(gridconfigfile, "w" ) as f:
@@ -518,8 +649,7 @@ def wassgridsurface_main():
     if path.exists( planefile ):
         print("Found! Loading...")
         all_planes = np.loadtxt( planefile )
-        meanplane = np.mean( all_planes, axis=0)
-
+        meanplane = np.nanmean( all_planes, axis=0)
 
 
     if not path.exists( args.outdir ):
@@ -579,6 +709,11 @@ def wassgridsurface_main():
 
         if args.gridsetup is None:
             print("Grid setup file not specified. See --gridsetup option")
+            sys.exit(-1)
+
+        if args.stereo_image_idx > 1 or args.stereo_image_idx < 0:
+            print("--stereo_image_idx must be 0 or 1")
+            sys.exit(-1)
 
         grid( wass_frames,
               matfile=args.gridsetup,
@@ -593,12 +728,15 @@ def wassgridsurface_main():
                   "MAX_ITERS": args.dct_maxiters,
                   "TOLERANCE_CHANGE": args.dct_maxtol,
                   "LEARNING_RATE": args.dct_lr,
-                  })
+                  },
+              NUM_PARALLEL_PROCESSES=args.parallel,
+              image_id_to_save=args.stereo_image_idx,
+              force_zero_mean=args.force_zero_mean)
 
         print("Gridding completed.")
 
     else:
-        print("Invalid actions specified.")
+        print("Invalid action specified.")
         sys.exit(-2)
 
 
